@@ -18,9 +18,6 @@
  *
  */
 
-#include "im.h"
-#include "safedelete.h"
-
 //! \class XMPP::Client client.h
 //! \brief Communicates with the XMPP network.  Start here.
 //!
@@ -69,34 +66,22 @@
 //!  }
 //!  \endcode
 
-#include <stdarg.h>
-#include <qobject.h>
+#include <QObject>
 #include <QMap>
-#include <qtimer.h>
-#include <qpointer.h>
-//Added by qt3to4:
+#include <QTimer>
+#include <QPointer>
 #include <QList>
+
+#include "im.h"
+#include "safedelete.h"
 #include "xmpp_tasks.h"
 #include "xmpp_xmlcommon.h"
 #include "s5b.h"
 #include "xmpp_ibb.h"
 #include "xmpp_bitsofbinary.h"
 #include "filetransfer.h"
-
-/*#include <stdio.h>
-#include <stdarg.h>
-#include <qstring.h>
-#include <qdom.h>
-#include <qobjectlist.h>
-#include <qtimer.h>
-#include "xmpp_stream.h"
-#include "xmpp_tasks.h"
-#include "xmpp_xmlcommon.h"
-#include "xmpp_dtcp.h"
-#include "xmpp_ibb.h"
-#include "xmpp_jidlink.h"
-
-using namespace Jabber;*/
+#include "xmpp_caps.h"
+#include "protocol.h"
 
 #ifdef Q_OS_WIN
 #define vsnprintf _vsnprintf
@@ -129,7 +114,8 @@ public:
 	int id_seed;
 	Task *root;
 	QString host, user, pass, resource;
-	QString osname, tzname, clientName, clientVersion, capsNode, capsVersion, capsExt;
+	QString osName, osVersion, tzname, clientName, clientVersion;
+	CapsSpec caps, serverCaps;
 	DiscoItem::Identity identity;
 	Features features;
 	QMap<QString,Features> extension_features;
@@ -139,6 +125,7 @@ public:
 
 	LiveRoster roster;
 	ResourceList resourceList;
+	CapsManager *capsman;
 	S5BManager *s5bman;
 	IBBManager *ibbman;
 	BoBManager *bobman;
@@ -155,12 +142,9 @@ Client::Client(QObject *par)
 	d->tzoffset = 0;
 	d->useTzoffset = false;
 	d->active = false;
-	d->osname = "N/A";
+	d->osName = "N/A";
 	d->clientName = "N/A";
 	d->clientVersion = "0.0";
-	d->capsNode = "";
-	d->capsVersion = "";
-	d->capsExt = "";
 
 	d->id_seed = 0xaaaa;
 	d->root = new Task(this, true);
@@ -174,10 +158,15 @@ Client::Client(QObject *par)
 	d->bobman = new BoBManager(this);
 
 	d->ftman = 0;
+
+	d->capsman = new CapsManager(this);
 }
 
 Client::~Client()
 {
+	//fprintf(stderr, "\tClient::~Client\n");
+	//fflush(stderr);
+
 	close(true);
 
 	delete d->ftman;
@@ -199,6 +188,7 @@ void Client::connectToServer(ClientStream *s, const Jid &j, bool auth)
 	//connect(d->stream, SIGNAL(closeFinished()), SLOT(streamCloseFinished()));
 	connect(d->stream, SIGNAL(incomingXml(QString)), SLOT(streamIncomingXml(QString)));
 	connect(d->stream, SIGNAL(outgoingXml(QString)), SLOT(streamOutgoingXml(QString)));
+	connect(d->stream, SIGNAL(haveUnhandledFeatures()), SLOT(parseUnhandledStreamFeatures()));
 
 	d->stream->connectToServer(j, auth);
 }
@@ -265,6 +255,11 @@ BoBManager *Client::bobManager() const
 	return d->bobman;
 }
 
+CapsManager *Client::capsManager() const
+{
+	return d->capsman;
+}
+
 bool Client::isActive() const
 {
 	return d->active;
@@ -408,14 +403,6 @@ QString Client::groupChatNick(const QString &host, const QString &room) const
 	return QString();
 }
 
-bool Client::isStreamManagementActive() const {
-	ClientStream *cs = qobject_cast<ClientStream*>(d->stream);
-	if(cs)
-		return cs->isStreamManagementActive();
-	return false;
-}
-
-
 /*void Client::start()
 {
 	if(d->stream->old()) {
@@ -431,6 +418,9 @@ bool Client::isStreamManagementActive() const {
 // TODO: fast close
 void Client::close(bool)
 {
+	//fprintf(stderr, "\tClient::close\n");
+	//fflush(stderr);
+
 	if(d->stream) {
 		d->stream->disconnect(this);
 		d->stream->close();
@@ -522,6 +512,9 @@ static QDomElement oldStyleNS(const QDomElement &e)
 
 void Client::streamReadyRead()
 {
+	//fprintf(stderr, "\tClientStream::streamReadyRead\n");
+	//fflush(stderr);
+
 	while(d->stream && d->stream->stanzaAvailable()) {
 		Stanza s = d->stream->read();
 
@@ -550,6 +543,19 @@ void Client::streamOutgoingXml(const QString &s)
 	emit xmlOutgoing(str);
 }
 
+void Client::parseUnhandledStreamFeatures()
+{
+	QList<QDomElement> nl = d->stream->unhandledFeatures();
+	foreach (const QDomElement &e, nl) {
+		if (e.localName() == "c" && e.namespaceURI() == NS_CAPS) {
+			d->serverCaps = CapsSpec::fromXml(e);
+			if (d->capsman->isEnabled()) {
+				d->capsman->updateCaps(Jid(d->stream->jid().domain()), d->serverCaps);
+			}
+		}
+	}
+}
+
 void Client::debug(const QString &str)
 {
 	emit debugText(str);
@@ -607,7 +613,7 @@ void Client::distribute(const QDomElement &x)
 	}
 }
 
-void Client::send(const QDomElement &x, bool want_notify)
+void Client::send(const QDomElement &x)
 {
 	if(!d->stream)
 		return;
@@ -626,13 +632,14 @@ void Client::send(const QDomElement &x, bool want_notify)
 		//printf("bad stanza??\n");
 		return;
 	}
+	emit stanzaElementOutgoing(e);
 	QString out = s.toString();
 	//qWarning() << "Out: " << out;
 	debug(QString("Client: outgoing: [\n%1]\n").arg(out));
 	emit xmlOutgoing(out);
 
 	//printf("x[%s] x2[%s] s[%s]\n", Stream::xmlToString(x).toLatin1(), Stream::xmlToString(e).toLatin1(), s.toString().toLatin1());
-	d->stream->write(s, want_notify);
+	d->stream->write(s);
 }
 
 void Client::send(const QString &str)
@@ -992,9 +999,9 @@ void Client::importRosterItem(const RosterItem &item)
 	debug(dstr + str);
 }
 
-void Client::sendMessage(const Message &m, bool want_notify)
+void Client::sendMessage(const Message &m)
 {
-	JT_Message *j = new JT_Message(rootTask(), m, want_notify);
+	JT_Message *j = new JT_Message(rootTask(), m);
 	j->go(true);
 }
 
@@ -1007,6 +1014,12 @@ void Client::sendSubscription(const Jid &jid, const QString &type, const QString
 
 void Client::setPresence(const Status &s)
 {
+	if (d->capsman->isEnabled()) {
+		if (d->caps.version().isEmpty() && !d->caps.node().isEmpty()) {
+			d->caps = CapsSpec(makeDiscoResult(d->caps.node())); /* recompute caps hash */
+		}
+	}
+
 	JT_Presence *j = new JT_Presence(rootTask());
 	j->pres(s);
 	j->go(true);
@@ -1020,7 +1033,12 @@ void Client::setPresence(const Status &s)
 
 QString Client::OSName() const
 {
-	return d->osname;
+	return d->osName;
+}
+
+QString Client::OSVersion() const
+{
+	return d->osVersion;
 }
 
 QString Client::timeZone() const
@@ -1058,24 +1076,24 @@ QString Client::clientVersion() const
 	return d->clientVersion;
 }
 
-QString Client::capsNode() const
+CapsSpec Client::caps() const
 {
-	return d->capsNode;
+	return d->caps;
 }
 
-QString Client::capsVersion() const
+CapsSpec Client::serverCaps() const
 {
-	return d->capsVersion;
-}
-
-QString Client::capsExt() const
-{
-	return d->capsExt;
+	return d->serverCaps;
 }
 
 void Client::setOSName(const QString &name)
 {
-	d->osname = name;
+	d->osName = name;
+}
+
+void Client::setOSVersion(const QString &version)
+{
+	d->osVersion = version;
 }
 
 void Client::setTimeZone(const QString &name, int offset)
@@ -1095,28 +1113,29 @@ void Client::setClientVersion(const QString &s)
 	d->clientVersion = s;
 }
 
-void Client::setCapsNode(const QString &s)
+void Client::setCaps(const CapsSpec &s)
 {
-	d->capsNode = s;
+	d->caps = s;
 }
 
-void Client::setCapsVersion(const QString &s)
-{
-	d->capsVersion = s;
-}
-
-DiscoItem::Identity Client::identity()
+DiscoItem::Identity Client::identity() const
 {
 	return d->identity;
 }
 
-void Client::setIdentity(DiscoItem::Identity identity)
+void Client::setIdentity(const DiscoItem::Identity &identity)
 {
+	if (!(d->identity == identity)) {
+		d->caps.resetVersion();
+	}
 	d->identity = identity;
 }
 
 void Client::setFeatures(const Features& f)
 {
+	if (!(d->features == f)) {
+		d->caps.resetVersion();
+	}
 	d->features = f;
 }
 
@@ -1125,30 +1144,78 @@ const Features& Client::features() const
 	return d->features;
 }
 
-void Client::addExtension(const QString& ext, const Features& features)
+DiscoItem Client::makeDiscoResult(const QString &node) const
 {
-	if (!ext.isEmpty()) {
-		d->extension_features[ext] = features;
-		d->capsExt = extensions().join(" ");
+	DiscoItem item;
+	item.setNode(node);
+	DiscoItem::Identity id = identity();
+	if (id.category.isEmpty() || id.type.isEmpty()) {
+		id.category = "client";
+		id.type = "pc";
 	}
-}
+	item.setIdentities(id);
 
-void Client::removeExtension(const QString& ext)
-{
-	if (d->extension_features.contains(ext)) {
-		d->extension_features.remove(ext);
-		d->capsExt = extensions().join(" ");
+	Features features;
+
+	if (d->ftman) {
+		features.addFeature("http://jabber.org/protocol/bytestreams");
+		features.addFeature("http://jabber.org/protocol/ibb");
+		features.addFeature("http://jabber.org/protocol/si");
+		features.addFeature("http://jabber.org/protocol/si/profile/file-transfer");
 	}
-}
+	features.addFeature("http://jabber.org/protocol/disco#info");
+	features.addFeature("jabber:x:data");
+	features.addFeature("urn:xmpp:bob");
+	features.addFeature("urn:xmpp:ping");
+	features.addFeature("urn:xmpp:time");
 
-QStringList Client::extensions() const
-{
-	return d->extension_features.keys();
-}
+	// Client-specific features
+	foreach (const QString & i, d->features.list()) {
+		features.addFeature(i);
+	}
 
-const Features& Client::extension(const QString& ext) const
-{
-	return d->extension_features[ext];
+	item.setFeatures(features);
+
+	// xep-0232 Software Information
+	XData si;
+	XData::FieldList si_fields;
+
+	XData::Field si_type_field;
+	si_type_field.setType(XData::Field::Field_Hidden);
+	si_type_field.setVar("FORM_TYPE");
+	si_type_field.setValue(QStringList(QLatin1String("urn:xmpp:dataforms:softwareinfo")));
+	si_fields.append(si_type_field);
+
+	XData::Field software_field;
+	software_field.setType(XData::Field::Field_TextSingle);
+	software_field.setVar("software");
+	software_field.setValue(QStringList(d->clientName));
+	si_fields.append(software_field);
+
+	XData::Field software_v_field;
+	software_v_field.setType(XData::Field::Field_TextSingle);
+	software_v_field.setVar("software_version");
+	software_v_field.setValue(QStringList(d->clientVersion));
+	si_fields.append(software_v_field);
+
+	XData::Field os_field;
+	os_field.setType(XData::Field::Field_TextSingle);
+	os_field.setVar("os");
+	os_field.setValue(QStringList(d->osName));
+	si_fields.append(os_field);
+
+	XData::Field os_v_field;
+	os_v_field.setType(XData::Field::Field_TextSingle);
+	os_v_field.setVar("os_version");
+	os_v_field.setValue(QStringList(d->osVersion));
+	si_fields.append(os_v_field);
+
+	si.setType(XData::Data_Result);
+	si.setFields(si_fields);
+
+	item.setExtensions(QList<XData>() << si);
+
+	return item;
 }
 
 void Client::s5b_incomingReady()
